#!/usr/bin/python3
#
#   Written by Christian D. Powell, 10/14/2019
#   Copyright (c) 2019. Christian D. Powell and Hunter N.B. Moseley
#   All rights reserved.
#
"""
Module for creating protein and alignment objects used in protamine analyses.

Separates entries from the uniprot_2019_05 xml file in Docs/Datasets/ into four homologous groups. The four groups are
eutherian sperm protamine P1, eutherian sperm protamine P2, metatehrain sperm protamine P1, and fish protamines. Only a
single entry is allowed per organism in a given group with preference given to Swiss-Prot (reviewed) entries. This goes
for all groups with the exception of the fish protamine group which is unfiltered.
"""
import jsonpickle
from ProtamineAnalysis import *


ABERRANT_P2_IDS = ["PRM2_BOVIN", "PRM2_PIG", "PRM2_RAT", "Q9EP54_CRIGR"]


if __name__ == '__main__':

    """
    Retrieving Main Dataset
    """
    with open("Docs/Datasets/uniprot_2019_05.xml", "r") as f:
        protamine_xml_str = f.read()
    xml_entries = uniprot.separate_entries(protamine_xml_str)
    protein_dicts = [uniprot.parse_xml_str(entry) for entry in xml_entries]

    """
    Formation of Homologous Protamine Groups
    """
    e_p1_list = list()
    e_p2_list = list()
    m_p1_list = list()
    fish_list = list()

    count = 0
    while count < len(protein_dicts):
        if 'Mammalia' in protein_dicts[count]['_OC']:  # changed from Eukaryota

            # Sort sperm protamine P1 types (eutherian and metatehrian)
            if any(name in protein_dicts[count]['_GN'] for name in ['PRM1', 'Prm1']):

                # Eutherian sperm protamine P1 type
                if 'Eutheria' in protein_dicts[count]['_OC']:
                    if protein_dicts[count]['_OX'] in [pro_dict['_OX'] for pro_dict in e_p1_list]:
                        for pro_dict in e_p1_list:
                            if pro_dict['_OX'] == protein_dicts[count]['_OX']:
                                if pro_dict['_DB'] == 'Swiss-Prot':
                                    break
                                else:
                                    pro_dict = protein_dicts[count]
                                    break
                    else:
                        e_p1_list.append(protein_dicts[count])

                # Metatherian sperm protamine P1 type
                elif 'Metatheria' in protein_dicts[count]['_OC']:
                    if protein_dicts[count]['_OX'] in [pro_dict['_OX'] for pro_dict in m_p1_list]:
                        for pro_dict in m_p1_list:
                            if pro_dict['_OX'] == protein_dicts[count]['_OX']:
                                if pro_dict['_DB'] == 'Swiss-Prot':
                                    break
                                else:
                                    pro_dict = protein_dicts[count]
                                    break
                    else:
                        m_p1_list.append(protein_dicts[count])

            # Eutherian sperm protamine P2 type
            elif any(name in protein_dicts[count]['_GN'] for name in ['PRM2', 'Prm2']):
                if 'Eutheria' in protein_dicts[count]['_OC']:
                    if not any(A_ID == protein_dicts[count]['_ID'] for A_ID in ABERRANT_P2_IDS):
                        if protein_dicts[count]['_OX'] in [pro_dict['_OX'] for pro_dict in e_p2_list]:
                            for pro_dict in e_p2_list:
                                if pro_dict['_OX'] == protein_dicts[count]['_OX']:
                                    if pro_dict['_DB'] == 'Swiss-Prot':
                                        break
                                    else:
                                        pro_dict = protein_dicts[count]
                                        break
                        else:
                            e_p2_list.append(protein_dicts[count])

        # Fish sperm protamines
        elif 'Actinopterygii' in protein_dicts[count]['_OC']:
            if 'like' not in protein_dicts[count]['_DE'] and \
                    protein_dicts[count]['_DE'] != '':  # removes protamine-like proteins
                fish_list.append(protein_dicts[count])

        count += 1

    """
    Protein Dictionaries to Protein Objects
    """
    e_p1_pro_list = [protein.Protein(**dict(pro_dict)) for pro_dict in e_p1_list]
    e_p2_pro_list = [protein.Protein(**dict(pro_dict)) for pro_dict in e_p2_list]
    m_p1_pro_list = [protein.Protein(**dict(pro_dict)) for pro_dict in m_p1_list]
    fish_pro_list = [protein.Protein(**dict(pro_dict)) for pro_dict in fish_list]

    """
    Alignment Creation
    """
    eutheria_p1_align = align.Alignment(e_p1_pro_list)
    eutheria_p2_align = align.Alignment(e_p2_pro_list)
    metatheria_p1_align = align.Alignment(m_p1_pro_list)
    fish_align = align.Alignment(fish_pro_list)

    with open('Docs/Datasets/eutherian_p1.json', 'w') as f:
        f.write(jsonpickle.encode(eutheria_p1_align))
    with open('Docs/Datasets/eutherian_p2.json', 'w') as f:
        f.write(jsonpickle.encode(eutheria_p2_align))
    with open('Docs/Datasets/metatherian.json', 'w') as f:
        f.write(jsonpickle.encode(metatheria_p1_align))
    with open('Docs/Datasets/fish.json', 'w') as f:
        f.write(jsonpickle.encode(fish_align))
