#!/usr/bin/python3
#
#   Written by Christian D. Powell, 10/14/2019
#   Copyright (c) 2019. Christian D. Powell and Hunter N.B. Moseley
#   All rights reserved.
#
"""
Module containing Alignment class. Alignment class is a structure creating, managing, and manipulating a Multiple
Sequence Alignment (MSA) of amino acid sequences.
"""
from math import log2
import ProtamineAnalysis.muscle_tools as muscle

"""
TODO: Add citation to were they are from.
Ala (A) 9.06   Gln (Q) 3.78   Leu (L) 9.87   Ser (S) 6.69
Arg (R) 5.69   Glu (E) 6.18   Lys (K) 5.01   Thr (T) 5.55
Asn (N) 3.90   Gly (G) 7.27   Met (M) 2.38   Trp (W) 1.29
Asp (D) 5.47   His (H) 2.19   Phe (F) 3.94   Tyr (Y) 2.93
Cys (C) 1.20   Ile (I) 5.73   Pro (P) 4.84   Val (V) 6.87
Asx (B) 0      Glx (Z) 0      Xaa (X) 0.05
"""
BACKGROUND_FREQ = {
    'A': 0.0906, 'Q': 0.0378, 'L': 0.0987, 'S': 0.0669,
    'R': 0.0569, 'E': 0.0618, 'K': 0.0501, 'T': 0.0555,
    'N': 0.0390, 'G': 0.0727, 'M': 0.0238, 'W': 0.0129,
    'D': 0.0547, 'H': 0.0219, 'F': 0.0394, 'Y': 0.0293,
    'C': 0.0120, 'I': 0.0573, 'P': 0.0484, 'V': 0.0687,
    'B': 0.0001, 'Z': 0.0001, 'X': 0.0005
}


class Alignment(object):
    """
    Class for creating, managing, and manipulating a Multiple Sequence Alignment (MSA) of amino acid sequences.
    """

    def __init__(self, protein_list, back_freq_dict=BACKGROUND_FREQ):
        """
        Constructor for Alignment class.

        :param protein_list: List of protein objects to be aligned.
        :type protein_list: list[ProtamineAnalysis.protein_tools.Protein]
        :param back_freq_dict: Dictionary of amino acid background frequencies.
        :type back_freq_dict: dict[str: float]
        """
        self.protein_list = protein_list
        self.back_freq_dict = back_freq_dict
        self.fasta_str = muscle.align("".join([pro.fasta for pro in self.protein_list]))
        self.id_list, self.alignment_matrix = self._alignment()
        self.count_dicts = self._count_dicts()
        self.freq_dicts = self._freq_dicts()
        self.relative_entropy = self._relative_entropy(self.back_freq_dict)

    def __len__(self):
        """
        Method for returning the length of the alignment (the number of proteins within the alignment).

        :return: Number of proteins in alignment.
        :rtype: int
        """
        return len(self.protein_list)

    def __iter__(self):
        """
        Method for returning a iterable version of the Alignment object.

        :return:
        """
        return zip(*self.alignment_matrix)

    def __str__(self):
        """
        Returns the consensus sequence for the alignment.

        :return: Consensus sequence for the alignment.
        :rtype: str
        """
        align_len = len(self.id_list)
        return "".join(
            [max(pos, key=pos.get) for i, pos in enumerate(self.freq_dicts) if
             int(self.count_dicts[i].get('-') or 0) / align_len <= 0.4]
        )

    def extend(self, protein_list):
        """
        Method for extending the alignment (appending additional protein sequences and recreating the alignment).

        :param protein_list: List of protein objects to be added to the alignment.
        :type protein_list: list[ProtamineAnalysis.protein_tools.Protein]
        """
        self.protein_list.extend(protein_list)
        self.fasta_str = muscle.align("".join([pro.fasta for pro in self.protein_list]))
        self.id_list, self.alignment_matrix = self._alignment()
        self.count_dicts = self._count_dicts()
        self.freq_dicts = self._freq_dicts()
        self.relative_entropy = self._relative_entropy(self.back_freq_dict)

    def _alignment(self):
        """
        Method for taking a FASTA file containing multiple aligned FASTA entries and parsing that into a dictionary with
        sequence identifiers acting as the dictionary keys and the aligned sequences (in list format) as the values.

        :returns List of protein ids and a nested list representing the alignment.
        :rtype tuple[list[str], list[list[str]]
        """
        fasta_list = ['>' + f_str for f_str in self.fasta_str.split('>')[1:]]
        identification_list = list()
        alignment_matrix = list()

        # loop through the FASTA strings in the list
        for curr_fstr in fasta_list:
            identification_list.append(curr_fstr[curr_fstr.rfind('|') + 1:curr_fstr.find(' ')])
            alignment_matrix.append(list(curr_fstr[curr_fstr.find('\n'):].replace('\n', '')))

        return identification_list, alignment_matrix

    def _count_dicts(self):
        """
        Method for generating a list of dictionaries of the represented amino acids in each position (key) and number of
        occurrences of the given amino acid within the position (value).

        Used in entropy calculations.

        :return: List of amino acid counts for each position in the alignment.
        :rtype: list[dict[str: int]]
        """
        return [{aa: pos.count(aa) for aa in pos} for pos in self]

    def _freq_dicts(self):
        """
        Method for generating a list of dictionaries of the represented amino acids in each position (key) and frequency
        of the given amino acid within the position (value).

        Used in entropy calculations.

        :return: List of amino acid frequencies for each position in the alignment.
        :rtype: list[dict[str: float]]
        """
        align_len = len(self.id_list)
        return [
            {aa: pos[aa] / (align_len - int(pos.get('-') or 0)) for aa in pos if aa is not '-'}
            for pos in self.count_dicts
        ]

    def _relative_entropy(self, back_freq=BACKGROUND_FREQ):
        """
        Method for calculating the relative entropy (Kullback–Leibler divergence) of each position within the alignment.

        Dkl(P||Q) = -sum(P(x)log2(Q(x)/P(x)))
        score = D(P||Q) * % ungapped

        :param back_freq: Dictionary of amino acid background frequencies (defaults to natural abundancies).
        :type back_freq: dict[str: float]
        :return: List of relative entropy scores for each position in the alignment.
        :rtype: list[float]
        """
        align_len = len(self.id_list)
        # removed "if aa is not '-'" as it is redundant
        entropy_list = [sum([pos[aa] * log2(pos[aa] / back_freq[aa]) for aa in pos]) for pos in self.freq_dicts]
        # gap weighting is just D(P||Q) * % ungapped
        entropy_list = [entropy * (align_len - int(self.count_dicts[i].get('-') or 0)) / align_len for i, entropy in
                        enumerate(entropy_list)]
        return entropy_list

    def print_alignment(self, line_len=80, start=0, end=None, highlight_type=None,
                        threshold=log2(1/BACKGROUND_FREQ['R']) + 0.001, ids=None):
        """
        Method for displaying a multiple sequence alignment (MSA) and visualizing conserved positions within the MSA.

        :param line_len: number of residues to be printed per line.
        :type line_len: int
        :param start: starting position to begin printing residues from.
        :type start: int
        :param end: end position to terminate printing residues at.
        :type end: int
        :param highlight_type: method for
        :param threshold: threshold for determining which positions in the alignment are conserved.
        :type threshold: float
        :param ids: list of protein ids to exclusively output.
        :type ids: list[str]
        :return:
        """
        # output the entire alignment if no end is specified
        if not end:
            end = len(self.alignment_matrix[0])

        if ids:
            id_indexes = [ind for ind, i in enumerate(self.id_list) if any(i2 in i for i2 in ids)]
        else:
            id_indexes = range(0, len(self.id_list))

        # outputs each block of the alignment (ie. line numbers, all ids, and all amino acids in positions
        for x in range(start, end, line_len):

            if x + line_len > end:
                line_end = end
            else:
                line_end = x + line_len

            # prints the line numbers at the top of each
            print(' ' * 16, end='')
            for y in range(x, round(line_end, -1), 10):
                print('{0: >11}'.format(str(y + 10)), end='')
            print()

            # print each entry
            for y in id_indexes:
                print('{0: <17}'.format(self.id_list[y]), end='')
                for z in range(x, x + line_end, 10):
                    for a in range(z, z + 10):
                        if a >= end:
                            break
                        if threshold:
                            if next(iter(sorted(self.freq_dicts[a], key=lambda k: k[0], reverse=True))) != 'M' and \
                                    self.relative_entropy[a] >= threshold:
                                print('\\textcolor{Green}{' + self.alignment_matrix[y][a] + '}', end='')
                                # print('\x1b[30;42m' + self.alignment_matrix[y][a] + '\x1b[0m', end='')
                            else:
                                print(self.alignment_matrix[y][a], end='')
                        else:
                            print(self.alignment_matrix[y][a], end='')
                    print(' ', end='')
                print()

    def print_alignment_2(self, line_len=80, start=0, end=None, highlight_type=None,
                    threshold=log2(1 / BACKGROUND_FREQ['R']) + 0.001, ids=None):
        """
        Method for displaying a multiple sequence alignment (MSA) and visualizing conserved positions within the MSA.

        :param line_len: number of residues to be printed per line.
        :type line_len: int
        :param start: starting position to begin printing residues from.
        :type start: int
        :param end: end position to terminate printing residues at.
        :type end: int
        :param highlight_type: method for
        :param threshold: threshold for determining which positions in the alignment are conserved.
        :type threshold: float
        :param ids: list of protein ids to exclusively output.
        :type ids: list[str]
        :return:
        """
        # output the entire alignment if no end is specified
        if not end:
            end = len(self.alignment_matrix[0])

        if ids:
            id_indexes = [ind for ind, i in enumerate(self.id_list) if any(i2 in i for i2 in ids)]
        else:
            id_indexes = range(0, len(self.id_list))

        # outputs each block of the alignment (ie. line numbers, all ids, and all amino acids in positions
        for x in range(start, end, line_len):

            # prevents printing past the end of the alignment (array index out of bounds)
            if x + line_len > end:
                line_end = end
            else:
                line_end = x + line_len

            # prints the line numbers at the top of each
            print(' ' * 16, end='')
            for y in range(x, round(line_end, -1), 10):
                print('{0: >11}'.format(str(y + 10)), end='')
            print()

            # print asterisks '*' above highly conserved positions.
            print(' ' * 17, end='')
            y = x
            while y < line_end:
                if self.relative_entropy[y] >= threshold and (y - start) % 10 == 9:
                    print('*', end=' ')
                elif (y - start) % 10 == 9:
                    print(' ', end=' ')
                elif self.relative_entropy[y] >= threshold:
                    print('*', end='')
                else:
                    print(' ', end='')
                y += 1
            print()

            # print each entry
            for y in id_indexes:
                print('{0: <17}'.format(self.id_list[y]), end='')
                for z in range(x, x + line_end, 10):
                    for a in range(z, z + 10):
                        if a < end:
                            print(self.alignment_matrix[y][a], end='')
                    print(' ', end='')
                print()
        print()

    def protein_index(self, protein_id, index):
        """
        Method for determining protein specific index of a position within the alignment.

        :param protein_id: Protein identifier string.
        :type protein_id: str
        :param index: Alignment position.
        :type index: int
        :return: Index of alignment position within specified protein.
        :rtype: int
        """
        # fetches the index of the protein in the alignment
        id_index = self.id_list.index(protein_id)
        # converts alignment index into protein index (removes gaps)
        pro_index = index - self.alignment_matrix[id_index][:index + 1].count('-')

        if pro_index >= 0:
            return pro_index
        else:  # protein is likely a fragment
            return None
