#!/usr/bin/python3
#
#   Written by Christian D. Powell, 10/14/2019
#   Copyright (c) 2019. Christian D. Powell and Hunter N.B. Moseley
#   All rights reserved.
#
"""
Module contains a number of functions used in the analysis of Protamine sequences.
"""
from copy import deepcopy
from numpy import log2, percentile, std

AA = (
    'A', 'Q', 'L', 'S',
    'E', 'K', 'T',
    'N', 'G', 'M', 'W',
    'D', 'H', 'F', 'Y',
    'C', 'I', 'P', 'V',
    'B', 'Z', 'X',
)


def frequency(protein, amino_acids, start=None, end=None):
    """
    Method for determining the frequency of given amino acids within a protein.

    :param protein: Protein object.
    :type protein: ProtamineAnalysis.protein_tools.Protein
    :param amino_acids: List of amino acid characters to be counted.
    :type amino_acids: list or char
    :param start: Index of position to begin counting from in protein's sequence.
    :type start: int
    :param end: Index of position to end counting at in protein's sequence.
    :type end: int

    :return: Frequency of given amino acids in the given protein's sequence.
    :rtype: float
    """
    length = -1
    if start and end:
        length = end - start
    elif start and not end:
        length = len(protein) - start
    elif not start and end:
        length = end
    else:
        length = len(protein)

    if type(amino_acids) != list:
        amino_acids = list(amino_acids)

    return sum([protein.sequence.count(aa, start, end) for aa in amino_acids]) / length


def region_frequencies(alignment, amino_acids, start=0, end=-1, debug=False):
    """
    Method for determining the frequencies of given amino acids in proteins of a given alignment within a specific
    region of the alignment.

    :param alignment: Alignment object.
    :type alignment: ProtamineAnalysis.alignment_tools.Alignment
    :param amino_acids: List of amino acid characters to be counted.
    :type amino_acids: list or char
    :param start: Index of alignment position to begin counting from.
    :type start: int
    :param end: Index of alignment position to end counting at.
    :type end: int

    :return: Returns list containing the frequency of the given amino acids within the designated region.
    :rtype: list[float]
    """
    charge_freq_list = list()
    for pro in alignment.protein_list:
        s = alignment.protein_index(pro.id, start)
        e = alignment.protein_index(pro.id, end)
        if debug:
            if pro.sequence[s: e] == '':
                print(pro.id, s, e)
            print(pro.id, pro.sequence[s: e])
        charge_freq_list.append(frequency(pro, amino_acids, s, e))

    return charge_freq_list


def ranking_table(alignment, threshold=log2(1/0.0569), start=0):
    """
    Method for generating a plaintext table showing highly conserved positions within a given alignment. Rank tables used in
    supplemental tables of manuscript.

    example:
        Eutherian P1 Sperm Protamines
        ┌──────┬──────────┬─────────┬────────────┐
        │Rank  │Position  │Relative │Represented │
        │      │in        │Entropy  │Residues    │
        │      │Alignment │(bit)    │            │
        ├──────┼──────────┼─────────┼────────────┤
        │     1│         7│   6.1571│  C:   98.6%│
        │      │          │         │  R:    1.4%│
        ├──────┼──────────┼─────────┼────────────┤
        │     2│        49│   6.1198│  C:  97.92%│
        │      │          │         │  G:   1.39%│
        │      │          │         │  S:   0.69%│
        ...


    threshold=log2(1/0.0569) = 4.135...
        is equivalent to a position consisting solely of arginine residues.

    :param alignment: Alignment for the table to be generated.
    :type alignment: ProtamineAnalysis.alignment_tools.Alignment
    :param threshold: Conservation score threshold.
    :type threshold: float
    :param start: Alignment position to begin ranking from.
    :type: start: int

    :return: Outputs (in command line) a LaTeX formatted string representing a ranking table.
    """
    positions = [(count, ele) for count, ele in enumerate(alignment.relative_entropy)]
    positions.sort(key=lambda x: x[1], reverse=True)

    print("┌──────┬──────────┬─────────┬────────────┐")
    print("│Rank  │Position  │Relative │Represented │")
    print("│      │in        │Entropy  │Residues    │")
    print("│      │Alignment │(bit)    │            │")

    # loop through and output conserved positions.
    x = 0  # sentinel value.
    rank = 1
    while True:
        if positions[x][1] > threshold:
            if positions[x][0] >= start:
                # Primary residue is NOT methionine.
                if next(iter(sorted(alignment.freq_dicts[positions[x][0]], key=lambda k: k[0], reverse=True))) != 'M':
                    # prints 1) rank, 2) position in alignment, 3) relative entropy score, and 4) represented residues.
                    print("├──────┼──────────┼─────────┼────────────┤")
                    print('│{:>6}│{:>10}│{:>9.5}│'.format(
                        rank,
                        positions[x][0] + 1,
                        positions[x][1]
                    ), end='')
                    for i, v in enumerate(sorted(alignment.freq_dicts[positions[x][0]].items(), key=lambda kv: kv[1], reverse=True)):
                        if i == 0:
                            print('{:>3}: {:>6}%│'.format(v[0], round(v[1] * 100, 2)))
                        else:
                            print('│      │          │         │{:>3}: {:>6}%│'.format(v[0], round(v[1] * 100, 2)))
                    rank += 1

            x += 1  # update sentinel.

        else:  # iterated through all positions above the threshold.
            break
    print("└──────┴──────────┴─────────┴────────────┘")
    print()


def ranking_table_latex(alignment, threshold=log2(1/0.0569), start=0):
    """
    Method for generating a LaTeX table showing highly conserved positions within a given alignment. Rank tables used in
    supplemental tables of manuscript.

    example. $ ranking_latex_table(eutherian_P2_alignment)
    Rank & Position in Alignment & Relative Entropy (bit) & Represented Residues \\
    \hline
    \hline
    1 & 59 & 6.381 & \begin{tabular}{c}C: 100.0\%\end{tabular} \\
    \hline
    ...

    threshold=log2(1/0.0569) = 4.135...
        is equivalent to a position consisting solely of arginine residues.

    :param alignment: Alignment for the table to be generated.
    :type alignment: ProtamineAnalysis.alignment_tools.Alignment
    :param threshold: Conservation score threshold.
    :type threshold: float

    :return: Outputs (in command line) a LaTeX formatted string representing a ranking table.
    """
    positions = [(count, ele) for count, ele in enumerate(alignment.relative_entropy)]
    positions.sort(key=lambda x: x[1], reverse=True)

    print('Rank & Position in Alignment & Relative Entropy (bit) & Represented Residues \\\\\n\\hline')

    # loop through and output conserved positions.
    x = 0  # sentinel value.
    rank = 1
    while True:
        if positions[x][1] > threshold:
            if positions[x][0] >= start:
                # Primary residue is NOT methionine.
                if next(iter(sorted(alignment.freq_dicts[positions[x][0]], key=lambda k: k[0], reverse=True))) != 'M':
                    # prints 1) rank, 2) position in alignment, 3) relative entropy score, and 4) represented residues.
                    print('\\hline\n' + ' & '.join(
                        [str(rank), str(positions[x][0] + 1), '{:0.4}'.format(positions[x][1]),
                         '\\begin{tabular}{c}' + ' \\\\ '.join(
                             ['{}: {:>5}\\%'.format(i[0], round(i[1] * 100, 1)) for i in
                              sorted(alignment.freq_dicts[positions[x][0]].items(), key=lambda kv: kv[1], reverse=True)]
                         )
                         + '\\end{tabular}']) + ' \\\\')
                    rank += 1

            x += 1  # update sentinel.

        else:  # iterated through all positions above the threshold.
            break
    print()


def quartile_table(frequency_dicts):
    """

    :param frequency_dicts:
    :return:
    """
    print("┌────────────────┬──────┬──────┬──────┬──────┬──────┐")
    print("│Name            │Min   │Q1    │Median│Q3    │Max   │")
    for k, v in frequency_dicts.items():
        print("├────────────────┼──────┼──────┼──────┼──────┼──────┤")
        print("│{:16}".format(k), end='')
        quartile_list = list(percentile(v, [0, 25, 50, 75, 100]))
        for v2 in quartile_list:
            print("│{:>6.3}".format(v2), end='')
        print("│")
    print("└────────────────┴──────┴──────┴──────┴──────┴──────┘")
    print()


def quartile_table_latex(frequency_dicts):
    """

    :param frequency_dicts:
    :return:
    """
    print("Alignment & Min & Q1 & Median & Q3 & Max \\\\")
    for k, v in frequency_dicts.items():
        print(k + " & " + " & ".join(str(round(x, 3)) for x in percentile(v, [0, 25, 50, 75, 100])) + ' \\\\')
    print()


def variability(alignment, start=0, end=-1):
    """
    Method for calculating the reduction in variabilty caused due to each additional amino acids inclusion in the
    frequency analysis.

    :param alignment: Alignment for the table to be generated.
    :type alignment: ProtamineAnalysis.alignment_tools.Alignment
    :param start:
    :type start: int
    :param end:
    :type end: int
    :return:
    """
    range_list = list()

    for aa in AA:
        new_freq = region_frequencies(alignment, ['R', aa], start, end)
        range_list.append([aa, new_freq])

    std_list = deepcopy(range_list)

    range_list.sort(key=lambda x: max(x[1]) - min(x[1]))
    std_list.sort(key=lambda x: std(x[1]))

    return range_list, std_list


def variability_table(alignment, start=0, end=-1):
    """
    Method for generating a plaintext table showing the reduction of variability caused by each amino acid being
    included in the frequency analyses.

    example:
        ┌──────┬──────┬──────┬──────┬──────┐
        │Rank  │AA    │Range │AA    │std   │
        ├──────┼──────┼──────┼──────┼──────┤
        │     1│     K│ 0.141│     K│0.0272│
        │     2│     H│ 0.185│     H│0.0351│
        │     3│     L│ 0.222│     L│0.0387│
        │     4│     T│ 0.228│     T│0.0411│
        │     5│     A│ 0.256│     E│0.0448│
        └──────┴──────┴──────┴──────┴──────┘

    :param alignment: Alignment for the table to be generated.
    :type alignment: ProtamineAnalysis.alignment_tools.Alignment
    :param start:
    :type start: int
    :param end:
    :type end: int
    :return:
    """

    range_list, std_list = variability(alignment, start, end)

    print("┌──────┬──────┬──────┬──────┬──────┐")
    print("│Rank  │AA    │Range │AA    │std   │")
    print("├──────┼──────┼──────┼──────┼──────┤")
    for x in range(5):
        print('│{:>6}│{:>6}│{:>6.3}│{:>6}│{:>6.3}│'.format(
            x + 1,
            range_list[x][0],
            max(range_list[x][1]) - min(range_list[x][1]),
            std_list[x][0],
            std(std_list[x][1])
        ))
    print("└──────┴──────┴──────┴──────┴──────┘")
    print()


def variability_table_latex(alignment, start=0, end=-1):
    """
    Method for generating a LaTeX table showing the reduction of variability caused by each amino acid being included
    in the frequency analyses.

    example:
        Rank & Range & & STD &\\
        \hline
        1  &  K  &  0.141  &  K  &  0.027\\
        2  &  H  &  0.185  &  H  &  0.035\\
        3  &  L  &  0.222  &  L  &  0.039\\
        4  &  T  &  0.228  &  T  &  0.041\\
        5  &  A  &  0.256  &  E  &  0.045\\

    :param alignment: Alignment for the table to be generated.
    :type alignment: ProtamineAnalysis.alignment_tools.Alignment
    :param start:
    :type start: int
    :param end:
    :type end: int
    :return:
    """

    range_list, std_list = variability(alignment, start, end)

    print("Rank & Range & & STD &\\\\\n\\hline")
    for x in range(5):
        print(x + 1, " & ", range_list[x][0], " & ", round(max(range_list[x][1]) - min(range_list[x][1]), 3), " & ",
              std_list[x][0], " & ", str(round(std(std_list[x][1]), 3)) + "\\\\")
    print()
