#!/usr/bin/python3
#
#   Written by Christian D. Powell, 10/14/2019
#   Copyright (c) 2019. Christian D. Powell and Hunter N.B. Moseley
#   All rights reserved.
#
"""
Module containing Protein class and relevant functions. Protein class is a structure creating, managing, and
manipulating amino acid sequences from the UniProt knowledgebase.
"""


class Protein(object):
    """
    Class for creating, managing, and manipulating amino acid sequences.
    """

    def __init__(self, **kwargs):
        """
        Constructor for Protein class.

        :param kwargs:
        """
        self._DB = kwargs.get('_DB')  # DataBase (str) Not a true UniProt line type
        self._ID = kwargs.get('_ID')  # IDentifier (str)
        self._AC = kwargs.get('_AC')  # ACcession number (list)
        self._DE = kwargs.get('_DE')  # DEscription (str)
        self._GN = kwargs.get('_GN')  # Gene Name (str)
        self._OS = kwargs.get('_OS')  # Organism Species (str)
        self._OC = kwargs.get('_OC')  # Organism Classification (list)
        self._OX = kwargs.get('_OX')  # Organism taxonomy cross-reference (int)
        self._SV = kwargs.get('_SV')  # Sequence Version (int)
        self._PE = kwargs.get('_PE')  # Protein Existence (int)
        self._DR = kwargs.get('_DR')  # Database cross-Reference (dict)
        self._FT = kwargs.get('_FT')  # Feature Table (dict)
        self._SQ = kwargs.get('_SQ')  # SeQuence (str)

    def __len__(self):
        """
        Method for returning the length of the amino acid sequence.

        :return: Number of amino acids in protein.
        :rtype: int
        """
        return len(self._SQ)

    def __str__(self):
        """
        Returns the FASTA string for the protein.

        :return: FASTA string for the protein.
        :rtype: str
        """
        fasta_str = '>un|'
        if self._DB == "Swiss-Prot":
            fasta_str = '>sp|'
        elif self._DB == "TrEMBL":
            fasta_str = '>tr|'
        fasta_str += self._AC[0] + '|' + self._ID

        if self._DE:
            fasta_str += ' ' + self._DE
        if self._OS:
            fasta_str += ' OS=' + self._OS
        if self._OX:
            fasta_str += ' OX=' + str(self._OX)
        if self._GN:
            fasta_str += ' GN=' + self._GN
        if self._PE:
            fasta_str += ' PE=' + str(self._PE)
        if self._SV:
            fasta_str += ' SV=' + str(self._SV)

        return fasta_str

    @property
    def database_cross_reference(self):
        return self._DR

    @property
    def database(self):
        return self._DB

    @property
    def fasta(self):
        return self._fasta()

    @property
    def id(self):
        return self._ID

    @property
    def organism_class(self):
        return self._OC

    @property
    def sequence(self):
        return self._SQ

    @property
    def taxonomy(self):
        return self._OC

    def _fasta(self):
        fasta_str = str(self) + '\n'
        count = 0
        while count < len(self._SQ):
            fasta_str += self._SQ[count: count+60] + '\n'
            count += 60
        fasta_str += self._SQ[count:]
        return fasta_str
