#!/usr/bin/python3
#
#   Written by Christian D. Powell, 10/14/2019
#   Copyright (c) 2019. Christian D. Powell and Hunter N.B. Moseley
#   All rights reserved.
#
"""
This module contains a number of functions and classes to deal with retrieving from and interacting with the UniProt
(SwissProt) knowledgebase.
"""
from urllib import error, request
from xml.etree import ElementTree


def download(file_format, **kwargs):
    """
    Method for requesting files from the UniProt Knowledge base.

    :param file_format: File extension to be downloaded.
    :type file_format: str
    :param kwargs: Possible search parameters

    :return: Retrieved string of the requested file.
    :rtype str
    """
    retrieved_str = None

    if not kwargs:
        query_str = "https://www.uniprot.org/uniprot/?format=%s" % file_format

    else:
        query_str = "https://www.uniprot.org/uniprot/?query="
        query_param_str = list(kwargs.keys())[0] + ':' + kwargs[list(kwargs.keys())[0]]

        for key in list(kwargs.keys())[1:]:
            query_param_str += ' ' + key + ':' + kwargs[key]

        # encoding for url
        query_param_str = query_param_str.replace(' ', '%20')
        query_param_str = query_param_str.replace('\"', '%22')

        if file_format == 'xml':
            query_param_str = query_param_str.replace('_', '%20')
            query_param_str = query_param_str.replace(':', '%3A')
        # query_param_str = query_param_str.replace(' ', '+')
        # query_param_str = query_param_str.replace(':', '%3A')
        # query_param_str = query_param_str.replace('[', '%5B')
        # query_param_str = query_param_str.replace(']', '%5D')
        # query_param_str = query_param_str.replace('(', '%28')
        # query_param_str = query_param_str.replace(')', '%29')

        query_str += query_param_str + '&format=%s' % file_format

    try:
        with request.urlopen(query_str) as req:
            retrieved_str = req.read().decode('utf-8')  # the retrieved data must be formatted into utf-8
    except error.HTTPError as err:
        print(err)

    return retrieved_str


def separate_entries(xml_str):
    """
    Method for separating out cml entries from a UniProt xml file string.

    Used to quickly parse entries so that the entire download need not be parsed.

    :param xml_str: String of a UniProt xml file.
    :type xml_str: str
    :return: List of separated UniProt xml entries.
    :rtype list[str]
    """
    entry_list = ['<entry' + entry for entry in xml_str.split('<entry')[1:]]
    entry_list[-1] = entry_list[-1][: entry_list[-1].find('</entry>')] + '</entry>'

    return entry_list


def parse_dr(elems):
    """
    Method for parsing UniProt Database cross-Reference (DR) elements.

    * 'DR' key contains a list of entries
        * Each entry is a dictionary with two values
            * "type" - resource abbreviation (ie. Gene Ontology database)
            * "id" - resource identifier (ie. GO:0008987)

        Example:
        {..., 'DR': [{"type": "GO", "id": "GO:0008987"}, ...], ...}
    :param elems: ElementTree consisting of DR elements parsed from a UniProt xml entry.
    :type elems: xml.etree.ElementTree
    :return: List of DR elements.
    :rtype: list[dict[str: dict[str: str]]]
    """
    return [e.attrib for e in elems]


def parse_ft(elems):
    """
    Method for parsing UniProt Database Feature (FT) elements. FT elements are annotated residues or regions within the
    protein sequence.

    * 'FT' key contains a list of entries
        * Each entry is a dictionary with two values
            * "type" - resource abbreviation (ie. Gene Ontology database)
            * "id" - resource identifier (ie. GO:0008987)

        Example:
        {..., 'FT': [{"description": "Interchain (with C-23)", "evidence": "2", "location": [6], "type": "disulfide
        bond"}, ...], ...}

    :param elems: ElementTree consisting of FT elements parsed from a UniProt xml entry.
    :type elems: xml.etree.ElementTree
    :return: List of FT elements.
    :rtype: list[dict[str: str]]
    """
    ft_list = list()

    for e in elems:
        e_dict = e.attrib
        children = e.find('location').getchildren()

        if children[0].tag == 'position':
            e_dict['location'] = [int(children[0].attrib['position'])]
        elif children[0].tag == 'begin':

            # some entries have unknown indexes
            if children[0].attrib == {'status': 'unknown'}:
                e_dict['location'] = [int(children[1].attrib['position'])]
            elif children[1].attrib == {'status': 'unknown'}:
                e_dict['location'] = [int(children[0].attrib['position'])]

            else:
                e_dict['location'] = range(int(children[0].attrib['position']),
                                           int(children[1].attrib['position']) + 1)

        ft_list.append(e_dict)

    return ft_list


def parse_xml_str(xml_str):
    """
    Method for parsing UniProt xml file entries which have been converted to strings.

    :note: add {http://uniprot.org/uniprot} as needed

    :param xml_str: String of UniProt xml file entry to be parsed.
    :type xml_str: str
    :return: Dictionary comprised of the elements of the UniProt entry.
    :rtype: dict[str: ?]
    """
    elem = ElementTree.fromstring(xml_str)
    protein_existence = {
        'evidence at protein level': 1,
        'evidence at transcript level': 2,
        'inferred from homology': 3,
        'predicted': 4,
        'uncertain': 5
    }
    pro_dict = dict()

    # single element parameters
    pro_dict['_DB'] = elem.attrib['dataset']
    pro_dict['_ID'] = elem.find('name').text
    pro_dict['_AC'] = [e.text for e in elem.findall('accession')]
    try:
        pro_dict['_DE'] = elem.find('protein') \
            .find('recommendedName') \
            .find('fullName').text
    except AttributeError:
        pro_dict['_DE'] = ""
    try:
        pro_dict['_GN'] = elem.find('gene').find('name').text
    except AttributeError:
        pro_dict['_GN'] = ""

    pro_dict['_OS'] = elem.find('organism').find('name').text

    pro_dict['_OC'] = [e.text for e in elem.find('organism').find('lineage').findall('taxon')]
    pro_dict['_OX'] = elem.find('organism').find('dbReference').attrib['id']
    pro_dict['_SV'] = int(elem.find('sequence').attrib['version'])
    pro_dict['_PE'] = protein_existence[elem.find('proteinExistence').attrib['type']]
    pro_dict['_SQ'] = elem.find('sequence').text.replace('\n', '')

    # requires helper functions to parse these elements (multiple entries)
    pro_dict['_DR'] = parse_dr(elem.findall('dbReference'))
    pro_dict['_FT'] = parse_ft(elem.findall('feature'))

    return pro_dict
